#load "spitree.cmo";;
#load "aconv.cmo";;
#load "spiparser.cmo";;
#load "spilexer.cmo";;
#load "axioms.cmo";;
#load "constraints.cmo";;
#load "constraintgen.cmo";;
#load "ssolver.cma";;
#load "alfpgen.cmo";;

open Alfp
open Axioms
open Alfpgen;;

let infile = "spisamples/match.spi" in
let filechan = open_in infile in
let lexbuf = Lexing.from_channel filechan in
let ast = Spiparser.process Spilexer.spi_tokens lexbuf in
let (ast_substituted, accum, substMap) = Aconv.proc_subst ast 0 Aconv.SMap.empty in
let freevars =
  Aconv.SMap.fold (fun _ name list ->
		     Constraints.UnType(name) :: list) substMap [] in

  print_endline "Alpha converted process:";
  print_endline ((Spitree.string_of_proc ast_substituted) ^ "\n");
  print_endline "Free names:";
  Aconv.SMap.iter (Printf.printf "%s: %s\n") substMap;
  print_endline "\n\nAxioms:";
  print_commented_conjunction axioms;

  let typeconstraints = Constraintgen.constraint_of_proc freevars ast_substituted in
  let messagesformula =
(*    print_endline "Beskeder:\n";
    List.iter (fun element ->
		 print_endline (Spitree.string_of_msg element)
	      ) Constraintgen.messages#get;*)
    let messages = Constraintgen.messages#get in
    let hd, tl = List.hd messages, List.tl messages in
    let firstname, firstformula = alfp_of_msg hd in
      List.fold_left (fun accum_cl element ->
			let propername, formula = alfp_of_msg element in
			  ClAnd(accum_cl, formula)
		     ) firstformula tl
  in
  let dom_conjunction = domformula Constraintgen.environments#get in
  let encodings = clause_of_formula typeconstraints in
  let envformula = formula_of_envlist Constraintgen.environments#get in
  let finalconjunction = (stratify (ClAnd(envformula,ClAnd(dom_conjunction, (ClAnd(axiom_conjunction, ClAnd(messagesformula, encodings))))))) in
    print_endline (string_of_cl ~commented:true encodings);
    print_endline ("Noncommented formula:\n\n" ^ (string_of_cl finalconjunction) ^ "\n\nStop noncommented formula");
    Ssolver.deducible_facts finalconjunction "ssolver/heap"
