/* spiparser.mly */
/* Spi Calculus parser */
%{
open Printf
let parse_error s = (* Called by the parser function on error *)
  print_endline ("spiparser: " ^ s);
  flush stdout;;
%}
%token <string> ID
%token <string> LVAR
%token <char> CHAR
%token NIL OK
%token BEGIN END BANG PAR OUT IN NEW DECRYPT SPLIT MATCH
%token CHAN UN KEY TOK TPAIR UNKNOWN
%token EMPTY PAIR COMMA LPAREN RPAREN LBRACE RBRACE COLON SCOLON AS DASH

%start process
%start message
%type <Spitree.proc> process
%type <Spitree.msg> message

%%
process:    NIL  { Spitree.PNil }
	  | OUT LPAREN message COMMA message RPAREN  { Spitree.Out($3, $5) }
	  | IN LPAREN message COMMA ID RPAREN SCOLON process  { Spitree.In($3, $5, $8) }
	  | BANG IN LPAREN message COMMA ID RPAREN SCOLON process  { Spitree.InRepl($4, $6, $9) }
	  | NEW ID SCOLON process  { Spitree.Nu($2, Spitree.Unknown, $4) }
	  | LPAREN process parrest { if $3 = Spitree.PNil then
				       $2
				     else
				       Spitree.PPara($2, $3) }
	  | BEGIN ID LPAREN message RPAREN  { Spitree.Begin($2, $4) }
	  | END ID LPAREN message RPAREN  { Spitree.End($2, $4) }
	  | DECRYPT message AS LBRACE ID RBRACE message SCOLON process  { Spitree.Decrypt($2, $5, Spitree.Unknown, $7, $9) }
	  | SPLIT message AS LPAREN ID COMMA ID RPAREN SCOLON process  { Spitree.Split($2, $5, Spitree.Unknown, $7, Spitree.Unknown, $10) }
	  | MATCH message AS LPAREN message COMMA ID RPAREN SCOLON process  { Spitree.Match($2, Spitree.Unknown, $5, $7, Spitree.Unknown, $10) }
;
parrest:    PAR process parrest { if $3 = Spitree.PNil then
				    $2
				  else
				    Spitree.PPara($2, $3) }
	  | RPAREN { Spitree.PNil }
;
message:    OK  { Spitree.Ok }
	  | LBRACE message RBRACE message  { Spitree.Encr($2, $4) }
	  | PAIR LPAREN message COMMA message RPAREN  { Spitree.MPair($3, $5) }
	  | ID  { Spitree.Name($1) }
;

%%
