/* outputparser.mly */
/* Parser for output of the Succinct Solver */
%{
open Printf
let parse_error s = (* Called by the parser function on error *)
  print_endline ("SSolver outputparser: " ^ s);
  flush stdout;;
%}
%token <string> ID
%token <int> NUMBER
%token <char> CHAR
%token RELATION SLASH COLON
%token COMMA LPAREN RPAREN ENDING

%start output
%type <Alfp.term list list Alfp.SMap.t> output
%type <Alfp.term list list> tuples
%type <Alfp.term list> tuple
%type <Alfp.term list> termlist
%type <Alfp.term> term
%%

output:   RELATION ID SLASH NUMBER COLON tuples output { Alfp.SMap.add $2 $6 $7 }
        | RELATION ID SLASH NUMBER COLON output { $6 }
        | RELATION ID SLASH NUMBER COLON tuples ENDING { Alfp.SMap.add $2 $6 Alfp.SMap.empty }
	| RELATION ID SLASH NUMBER COLON ENDING {Alfp.SMap.empty};
tuples:   tuple tuples { $1::$2 }
	| tuple { [$1] };
tuple:    LPAREN termlist RPAREN COMMA { $2 };
termlist: term COMMA termlist { $1::$3 }
	| term { [$1] };
term:     ID { Alfp.Const($1) }
	| ID LPAREN termlist RPAREN { Alfp.FuncApp($1,$3) };
